@@ -25,7 +25,7 @@ from requests.utils import requote_uri
 from distutils.util import strtobool
 from pkg_resources import parse_version, get_distribution, DistributionNotFound
 
-__version__ = "1.6.0"
+__version__ = "1.6.1"
 __bagit_version__ = "1.8.1"
 __bagit_profile_version__ = "1.3.1"
 
@@ -37,18 +37,21 @@ else:  # pragma: no cover
     from urlparse import urlsplit, urlunsplit, urlparse
 
 try:
-    VERSION = get_distribution("bdbag").version
+    version = get_distribution("bdbag").version
+    VERSION = version + '' if not getattr(sys, 'frozen', False) else version + '-frozen'
 except DistributionNotFound:  # pragma: no cover
     VERSION = __version__ + '-dev' if not getattr(sys, 'frozen', False) else __version__ + '-frozen'
 PROJECT_URL = 'https://github.com/fair-research/bdbag'
 
 try:
-    BAGIT_VERSION = get_distribution("bagit").version
+    version = get_distribution("bagit").version
+    BAGIT_VERSION = version + '' if not getattr(sys, 'frozen', False) else version + '-frozen'
 except DistributionNotFound:  # pragma: no cover
     BAGIT_VERSION = 'unknown' if not getattr(sys, 'frozen', False) else __bagit_version__ + '-frozen'
 
 try:
-    BAGIT_PROFILE_VERSION = get_distribution("bagit_profile").version
+    version = get_distribution("bagit_profile").version
+    BAGIT_PROFILE_VERSION = version + '' if not getattr(sys, 'frozen', False) else version + '-frozen'
 except DistributionNotFound:  # pragma: no cover
     BAGIT_PROFILE_VERSION = 'unknown' if not getattr(sys, 'frozen', False) else __bagit_profile_version__ + '-frozen'
 
@@ -57,11 +60,12 @@ BDBAG_PROFILE_ID = 'https://raw.githubusercontent.com/fair-research/bdbag/master
 BDBAG_RO_PROFILE_ID = 'https://raw.githubusercontent.com/fair-research/bdbag/master/profiles/bdbag-ro-profile.json'
 
 CONTENT_DISP_REGEX = re.compile(r"^filename[*]=UTF-8''(?P<name>[-_.~A-Za-z0-9%]+)$")
-FILTER_REGEX = re.compile(r"(?P<column>^.*)(?P<operator>==|!=|=\*|!\*|\^\*|\$\*|>=|>|<=|<)(?P<value>.*$)")
+FILTER_REGEX = re.compile(r"(?P<column>^.*)(?P<operator>==|!=|=\*|!\*|=~|\^\*|\$\*|>=|>|<=|<)(?P<value>.*$)")
 FILTER_DOCSTRING = "\"==\" (equal), " \
                    "\"!=\" (not equal), " \
                    "\"=*\" (wildcard substring equal), " \
                    "\"!*\" (wildcard substring not equal), " \
+                   "\"=~\" (matches regular expression), " \
                    "\"^*\" (wildcard starts with), " \
                    "\"$*\" (wildcard ends with), " \
                    "or \">\", \">=\", \"<\", \"<=\""
@@ -150,13 +154,15 @@ def filter_dict(expr, entry):
     filter_val = expr_dict["value"]
     operator = expr_dict["operator"]
 
-    filter_neg = filter_substring = filter_relation = filter_startswith = filter_endswith = False
+    filter_neg = filter_substring = filter_re = filter_relation = filter_startswith = filter_endswith = False
     if "==" == operator:
         pass
     elif "!=" == operator:
         filter_neg = True
     elif "=*" == operator:
         filter_substring = True
+    elif "=~" == operator:
+        filter_re = True
     elif "^*" == operator:
         filter_startswith = True
     elif "$*" == operator:
@@ -182,6 +188,8 @@ def filter_dict(expr, entry):
         else:
             if filter_substring:
                 result = filter_val in str(value)
+            elif filter_re:
+                result = re.search(filter_val, str(value)) is not None
             elif filter_startswith:
                 result = str(value).startswith(filter_val)
             elif filter_endswith:
